# Copyright (c) 2002  Riverbank Computing Limited <info@riverbankcomputing.co.uk>
# Copyright (c) 2002  Jim Bublitz <jbublitz@nwinternet.com>

# This is the build script for PyKDE.  It should be run in the top level
# directory of the source distribution and by the Python interpreter for which
# it is being built.  It uses either qmake or tmake to do the hard work of
# generating the platform specific Makefiles.


import sys
import os
import glob
import re
import tempfile
import shutil
import py_compile
import string
import time

# Release numbers - don't forget to update these !!!
Release    = "3.7"
Version    = "3.7-4"

internalError = \
"""An internal error occured.  Please report all the output from the program,
including the following traceback, to <jbublitz@nwinternet.com> and CC the
PyKDE mailing list: <pykde@mats.imk.fraunhofer.de>. Thanks.
"""

# Define the globals.
progName        = None
qtDir           = None
qtIncDir        = None
qtVersion       = 0
qtLib           = None
qtLibDir        = None

kdeDir          = None
kdeIncDir       = None
kdeVersion      = 0
kdeLib          = None
pyQtSipDir      = None

platPyScriptDir = None
platPyDLLDir    = None
platPySiteDir   = None
platPyIncDir    = None
platPyLibDir    = None
platBinDir      = None
platMake        = None
platCopy        = None
platMove        = None
platRemove      = None
platMkdir       = None
platQTDIRName   = None
platKDEDIRName  = None

sipModuleDir    = None
sipIncDir       = None
sipBin          = None

modDir          = None
libDir          = "libs"
pyFullVers      = None
makefileGen     = None
makeBin         = None
proPatches      = {}
makefilePatches = {}
tempBuildDir    = None
enableOldPython = 0
catCppFiles     = 1
catSplit        = 1
nrThreads       = 0

FALSE      = 0
TRUE       = not FALSE

# QtVers must match qttags
QtVers     = ["300", "301", "302", "304", "305", "306", "310", "311", "312", "320"]
KDEVers    = ["300", "301", "303", "310", "311", "312", "313"]

modList    = {}
modList ["30"] = ["dcop", "kdecore", "kdesu", "kdefx", "kdeui", "kio", "kfile", "kparts", "khtml", "kjs", "kspell", "kdeprint"]
modList ["31"] = modList ["30"]

# Want this global so it's testable by external program
qttags = {
    0x020000: "Qt_1_43",
    0x020100: "Qt_2_00",
    0x020200: "Qt_2_1_0",
    0x020300: "Qt_2_2_0",
    0x020301: "Qt_2_3_0",
    0x030000: "Qt_2_3_1",
    0x030001: "Qt_3_0_0",
    0x030002: "Qt_3_0_1",
    0x030004: "Qt_3_0_2",
    0x030005: "Qt_3_0_4",
    0x030006: "Qt_3_0_5",
    0x030100: "Qt_3_0_6",
    0x030101: "Qt_3_1_0",
    0x030102: "Qt_3_1_1",
    0x030200: "Qt_3_1_2",
    0x040000: "Qt_3_2_0"
}


def usage(rcode = 2):
    """Display a usage message and exit.

    rcode is the return code passed back to the calling process.
    """
    print "Usage:"
    print "    %s [-h] " % (progName)
    print "    %s [options]\n" % (progName)
    print "where:"
    print "    -h             display this help message\n"

    print "  options:\n"
    print "    -c+            concatenate each module's C++ source files (default - use -c- to disable)"
    print "    -d dir         where PyKDE will be installed [default %s]" % (platPySiteDir)
    print "    -e dir         the directory containing the SIP header files [default %s]" % (sipIncDir)
    print "    -g prog        the name of the Makefile generator"
    print "    -i dir         the directory containing the Qt header files [default %s%sinclude]" % (platQTDIRName,os.sep)
    print "    -j #           split the concatenated C++ source files into # pieces [default 1]"
    print "    -k dir         the KDE base directory [default %s]" % (platKDEDIRName)
    print "    -l Qt-library  explicitly specify the type of Qt library (only qt-mt will work with PyKDE/KDE3)"
    print "    -m prog        the name of the Make program [default %s]" % (platMake)
    print "    -n dir         the directory containing the KDE header files [default %s/include]" % (platKDEDIRName)
    print "    -o dir         the directory containing the KDE lib files [default %s/lib]" % (platKDEDIRName)
    print "    -p dir         the name of the SIP code generator [default sip]"
    print "    -q dir         the Qt base directory [default %s]" % (platQTDIRName)
    print "    -r dir         the directory containing the Qt library [default %s%slib]" % (platQTDIRName,os.sep)
    print "    -s dir         the directory containing the SIP module"
    print "    -t dir         the directory containing the Python library directories [default %s]\n" % (platPyLibDir)
    print "    -v dir         the directory containing the PyQt sip files [default %s]" % (pyQtSipDir)
    print "    -z file        the name of a file containing command line flags"

    sys.exit(rcode)


def format(mess, err = FALSE):
    """Format a message by inserting line breaks at appropriate places.

    mess is the text of the message.

    Return the formatted message.
    """
    msg = ''
    margin = 78
    curs = 0
    if err:
        msg = "*" * margin + "\n"

    for w in string.split(mess):
        l = len(w)
        if curs != 0 and curs + l > margin:
            msg = msg + "\n"
            curs = 0

        if curs > 0:
            msg = msg + " "
            curs = curs + 1

        msg = msg + w
        curs = curs + l

    if err:
        msg = msg + "\n" + ("*" * margin) + "\n"
    else:
        msg = msg + '\n'

    return msg


def error(mess):
    """Display an error message and exit.

    mess is the text of the message.
    """
    frame = sys._getframe (1)
    ln = str (frame.f_lineno)

    sys.stderr.write(format ("Error: %s -- from line %s in build.py" % (mess, ln), TRUE))
    sys.exit(1)


def inform(mess):
    """Display an information message.

    mess is the text of the message.
    """
    sys.stdout.write(format(mess))
    sys.stdout.flush()


def makeVersion (s, v):
        ver = s + '_' + v [0] + '_' + v [1] + '_' + v [2]
        return ver


def initGlobals():
    """Sets the values of globals that need more than a simple assignment.
    This is called before the command line is parsed to set defaults for the
    help, and afterwards to reflect the command line.
    """
    global pyFullVers, pyVersNr

    pyFullVers = string.split(sys.version)[0]

    vl = string.split(re.findall("[0-9.]*",pyFullVers)[0],".")
    major = vl[0]
    minor = vl[1]

    pyVers = major + "." + minor
    pyVersNr = int(major) * 10 + int(minor)

    global platMake, platCopy, platMove, platRemove, platMkdir, platRmdir
    global platPyLibDir, platPyScriptDir, platPyDLLDir, platPySiteDir
    global platPyIncDir, platQTDIRName, platKDEDIRName, platBinDir

    if sys.platform == "win32":
        error ("KDE not supported under Windows")
    else:
        if platMake is None:
            platMake = "make"

        platCopy   = "cp"
        platMove   = "mv"
        platRemove = "rm -f"
        platMkdir  = "mkdir -p"
        platRmdir  = "rm -rf"

        if platPyLibDir is None:
            platPyLibDir = sys.prefix + "/lib/python" + pyVers

        platPyScriptDir = platPyLibDir
        platPyDLLDir = platPyLibDir + "/lib-dynload"

        if pyVersNr < 20:
            platPySiteDir = platPyLibDir
        else:
            platPySiteDir = platPyLibDir + "/site-packages"

        platPyIncDir = sys.prefix + "/include/python" + pyVers
        platQTDIRName = "$QTDIR"
        platKDEDIRName = "$KDEDIR"

        if platBinDir is None:
            platBinDir = "/usr/local/bin"

    global modDir, sipIncDir

    if modDir is None:
        modDir = platPySiteDir

    if sipIncDir is None:
        sipIncDir = platPyIncDir


def escape(s):
    """Escape all the backslashes (usually Windows directory separators) in a
    string so that they don't confuse regular expression substitutions.  Also
    enclose in double quotes if the string contains spaces.

    s is the string to escape.

    Returns the escaped string.
    """
    s = string.replace(s,"\\","\\\\")

    if string.find(s," ") >= 0:
        s = "\"" + s + "\""

    return s


def isProgram(f):
    """Return non-zero if a filename refers to an executable file.
    """
    return os.access(f,os.X_OK) and not os.path.isdir(f)


def searchPath(prog,xtradir = None):
    """Search the PATH for a program and return it's absolute pathname.

    prog is the name of the program.
    xtradir is the name of an optional extra directory to search as well.
    """
    # Create the list of directories to search.
    try:
        path = os.environ["PATH"]
    except KeyError:
        path = os.defpath

    dlist = string.split(path,os.pathsep)

    if xtradir:
        dlist.insert(0,xtradir)

    # Create the list of files to check.
    flist = []

    for d in dlist:
        flist.append(os.path.abspath(os.path.join(d,prog)))

    # Search for the file.
    for f in flist:
        if isProgram(f):
            return f

    # Nothing found.
    return None


def patchFile(name,pdict,bakfile = None):
    """Apply the patches to a file.

    name is the name of the file.
    pdict is the patch dictionary.
    bakfile is the name of the optional backup file for the unpatched version.
    """
    newname = name + ".new"

    fi = open(name,"r")
    fo = open(newname,"w")

    for line in fi.readlines():
        for (pattern, repl) in pdict.values():
            line = re.sub(pattern,repl,line)

        fo.write(line)

    fi.close()
    fo.close()

    if bakfile is None:
        os.remove(name)
    else:
        os.rename(name,bakfile)

    os.rename(newname,name)


def isQtLibrary(dir,lib):
    """See if a Qt library appears in a particular directory.

    dir is the name of the directory.
    lib is the name of the library.

    Returns the basename of the library file that was found or None if nothing
    was found.
    """

    lpatt = "/lib%s*" % (lib)

    l = glob.glob(dir + lpatt)

    if len(l) == 0:
        return None

    return os.path.basename(l[0])


def copyToFile(name,text):
    """Copy a string to a file.

    name is the name of the file.
    text is the contents to copy to the file.
    """
    f = open(name,"w")
    f.write(text)
    f.close()


def mkTempBuildDir(olddir = None):
    """Create a temporary build directory for a console application called
    qttest, complete with patched Makefile.  The global tempBuildDir is set to
    the name of the directory.  The temporary directory becomes the current
    directory.

    olddir is None if the directory should be created, otherwise it is deleted.

    Returns the name of the previous current directory.
    """
    global tempBuildDir

    if olddir is None:
        tempBuildDir = tempfile.mktemp()

        try:
            os.mkdir(tempBuildDir)
        except:
            error("Unable to create temporary directory")

        newdir = pushDir(tempBuildDir)

        copyToFile("qttest.pro",
"""TEMPLATE = app
TARGET = qttest
CONFIG = qt console warn_off release @BL_THREAD@
INCLUDEPATH = @BL_INCLUDEPATH@
DEFINES = @BL_DEFINES@
SOURCES = qttest.cpp

""")

        # Create a dummy source file to suppress a qmake warning.
        copyToFile("qttest.cpp","")
        buildMakefile("qttest.pro")

    else:
        popDir(olddir)
        newdir = None

        shutil.rmtree(tempBuildDir,1)

    return newdir


def checkQtDirAndVersion():
    """Check that Qt can be found and what its version is.
    """
    # Check the Qt directory is known and valid.
    global qtVersion, qtDir

    if qtDir is None:
        try:
            qtDir = os.environ["QTDIR"]
        except KeyError:
            error("Please set the name of the Qt base directory, either by using the -q argument or setting the QTDIR environment variable")
    else:
        os.environ["QTDIR"] = qtDir

    # For lib template project files.
    makefilePatches["QTDIR_LIB"] = [re.compile("(^CC)",re.M), "QTDIR    = " + qtDir + "\n\\1"]

    # For subdir template project files.
    makefilePatches["QTDIR_SUBDIR"] = [re.compile("(^MAKEFILE)",re.M), "QTDIR    = " + qtDir + "\n\\1"]

    if not os.access(qtDir,os.F_OK):
        error("The %s Qt base directory does not seem to exist. Use the -q argument or the QTDIR environment variable to set the correct directory" % (qtDir))

    inform("%s is the Qt base directory." % (qtDir))

    # Check that the Qt header files have been installed.
    if qtIncDir:
        qtincdir = qtIncDir
    else:
        qtincdir = os.path.join (qtDir,"include")

    qglobal = os.path.join (qtincdir, "qglobal.h")

    if not os.access(qglobal,os.F_OK):
        error("qglobal.h could not be found in %s. Use the -i argument to explicitly specify the correct directory" % (qtincdir))

    inform("%s contains qglobal.h." % (qtincdir))

    # Set the list of directories containing header files.
    incdirs = ". " + platPyIncDir

    if sipIncDir != platPyIncDir:
        incdirs = incdirs + " " + sipIncDir

    if qtIncDir:
        incdirs = incdirs + " " + qtIncDir

    proPatches["INCLUDEPATH"] = [re.compile("@BL_INCLUDEPATH@",re.M), incdirs]

    # Get the Qt version number.
    qtversstr = ""
    f = open(qglobal)
    l = f.readline()

    while l:
        wl = string.split(l)
        if len(wl) == 3 and wl[0] == "#define":
            if wl[1] == "QT_VERSION":
                qtv = wl[2]

                if qtv[0:2] == "0x":
                    qtVersion = int(qtv,16)
                else:
                    dec = int(qtv)
                    maj = dec / 100
                    min = (dec % 100) / 10
                    bug = (dec % 10)
                    qtVersion = (maj << 16) + (min << 8) + bug

            if wl[1] == "QT_VERSION_STR":
                qtversstr = wl[2][1:-1]

            if qtversstr != "" and qtVersion > 0:
                break

        l = f.readline()

    f.close()

    if qtversstr == "":
        error("The Qt version number could not be determined by parsing %s" % (qglobal))

    inform("Qt %s is being used." % (qtversstr))


def checkKDEIncl (path):
    if os.access (os.path.join (path, "kapplication.h"), os.F_OK):
        return path
    else:
        return ""


def checkKDEDirAndVersion():
    """Check that KDE can be found and what its version is.
    """

    # Check the KDE directory is known and valid.
    global kdeIncDir, kdeVersion, kdeDir

    if kdeDir is None:
        try:
            kdeDir = os.environ["KDEDIR"]
            kdeIncDir = checkKDEIncl (os.path.join (kdeDir, "include"))
            if not kdeIncDir:
                kdeIncDir = checkKDEIncl (os.path.join (kdeDir, "include", "kde"))

        except KeyError:
            pass

        if not kdeIncDir:
            inclCandidates = ["/opt/kde3/include", "/usr/include/kde", "/usr/include"]
            for path in inclCandidates:
                if checkKDEIncl (path) != "":
                    kdeIncDir = path
                    break

    else:
        kdeIncDir = checkKDEIncl (os.path.join (kdeDir, "include"))
        if not kdeIncDir:
            kdeIncDir = os.path.join (kdeDir, "include", "kde")

    if kdeDir is None and kdeIncDir:
        last = string.find (kdeIncDir, "/include")
        if last > 0:
            kdeDir = kdeIncDir [:last]
            os.environ["KDEDIR"] = kdeDir

    if not kdeIncDir:
        error("Please set the name of the KDE include directory, either by using the -t argument or setting the KDEDIR environment variable")

    # For lib template project files.
    makefilePatches["KDEDIR_LIB"] = [re.compile("(^CC)",re.M), "KDEDIR    = " + kdeDir + "\n\\1"]

    # For subdir template project files.
    makefilePatches["KDEDIR_SUBDIR"] = [re.compile("(^MAKEFILE)",re.M), "KDEDIR    = " + kdeDir + "\n\\1"]

    if not os.access(kdeDir,os.F_OK):
        error("The %s KDE base directory does not seem to exist. Use the -k argument or the KDEDIR environment variable to set the correct directory" % (kdeDir))

    inform("%s is the KDE base directory." % (kdeDir))

    # Check that the KDE header files have been installed.

    if kdeIncDir:
        kdeincdir = kdeIncDir
    else:
        kdeincdir = os.path.join (kdeDir, "include")

    # Get the KDE version number.

    kdeversstr = ""

    # Version info: KDE 3 - kdeversion.h; KDE 2 - kapp.h
    if os.access (os.path.join (kdeincdir, 'kdeversion.h'), os.F_OK):
        inform("%s contains kdeversion.h." % (kdeincdir))
        f = open(kdeincdir + '/kdeversion.h')
        l = f.readline()

        # Use MAJOR/MINOR/RELEASE values because KDE_VERSION and
        # KDE_VERSION_STRING are wrong in some distributions

        major = ''
        minor = ''
        rel   = ''

        while l:
            wl = string.split(l)
            if len(wl) == 3 and wl[0] == "#define":
                if wl[1] == "KDE_VERSION_MAJOR":
                    major = string.strip (wl[2])

                if wl[1] == "KDE_VERSION_MINOR":
                    minor = string.strip (wl[2])

                if wl[1] == "KDE_VERSION_RELEASE":
                    rel = string.strip (wl[2])

            if (major != '') and (minor != '') and (rel != ''):
                break

            l = f.readline()

        f.close()

        if (major == '') or (minor == '') or (rel == ''):
            error ("Couldn't read version info from kdeversion.h in %s" % (kdeincdir))

        kdeversstr = string.join ([major, minor, rel], '.')
        kdeVersion = int (major + minor + rel)

    else:
        error("kdeversion.h could not be found in %s. Use the -i argument to explicitly specify the correct directory" % (kdeincdir))

    kv = str (kdeVersion)
    if not kv in KDEVers:
        kdeVersion = 0
        for i in range (len (KDEVers) - 1, -1, -1):
            if KDEVers [i] < kv:
                kv = KDEVers [i]
                kdeVersion = int (kv)
                break

        if kdeVersion != 0:
            kdeversstr = "%s.%s.%s" % (kv [0], kv [1], kv [2])

    proPatches ["INCLUDEPATH"][1] = proPatches ["INCLUDEPATH"][1] + " " + kdeincdir
    proPatches ["KDE_INCLUDES"] = [re.compile("@KDE_INCLUDEPATH@",re.M), kdeincdir]

    if kdeversstr == "":
        error("The KDE version number could not be determined")

    inform("KDE %s is being used." % (kdeversstr))



def checkQtLibrary():
    """Check which Qt library is to be used.
    """
    global qtLib

    if qtLibDir:
        qtlibdir = qtLibDir
    else:
        qtlibdir = os.path.join (qtDir, "lib")


    qtLib = isQtLibrary(qtlibdir,"qt-mt")

    if qtLib:
        inform("The %s Qt library was found." % (qtLib))
    else:
        error("The %s Qt library could not be found in %s" % (qtLib,qtlibdir))

    proPatches["THREAD"] = [re.compile("@BL_THREAD@",re.M), "thread"]
    inform("Qt thread support is enabled.")


def checkThreading ():
    qtmodlib = platPySiteDir + '/libqtcmodule.so'
    os.system ('ldd %s > lddtmp' % (qtmodlib))
    f = open ('lddtmp', 'r')

    if re.search (re.compile("libqt-mt\.so" ,re.M), f.read ()) is None:
        inform ("Can't determine if PyQt uses libqt-mt")
        error ("PyQt linked to Qt lib which is incompatible with KDE version (threading)")

    else:
        inform ("Qt and KDE use compatible libs (threading)")


    f.close ()
    os.remove ('lddtmp')


def checkKDELibrary():
    """Check which Qt library is being used.
    """
    if kdeLib is None:
        kdelibdir = os.path.join (kdeDir, "lib")
    else:
        kdelibdir = kdeLib

    ml = modList [str (kdeVersion) [0:2]]

    print '\nfound libs for modules:'
    print '    ',

    for lib in ml:
        if lib == 'dcop':
            libpath = kdelibdir + '/libDCOP.so'
        elif (lib == 'kfile'):
            print 'kfile (kio)',
            continue
        else:
            libpath = kdelibdir + '/lib' + lib + '.so'

        if not os.access(libpath, os.F_OK):
            print
            error ("%s not found" % libpath)
        else:
            print lib,

    proPatches["KDELIBS"] = [re.compile("@BL_KDEDIR@",re.M), kdelibdir]

    print
    print


def checkPyQt ():
    """ Get the PyQt version info and locate the PyQt/sip directory
    """
    global pyQtSipDir

    if not pyQtSipDir:

        # report the actual PyQt version (NOT Qt version)
        # form the name of the PyQt top level directory
        try:
            from qt import PYQT_VERSION, PYQT_VERSION_STR
        except:
            error ("Couldn't import qt module from PyQt")

        pyqtv = hexVersionToString (hex(PYQT_VERSION))

        if pyqtv == Release:
            inform ("PyQt version is compatible with PyKDE-%s" % Release)
        else:
            error ("PyQt %s is not compativle with PyKDE %s" % (pyqtv, Release))

        pyqt = 'PyQt-x11-gpl-' + str (PYQT_VERSION_STR)

        # search these paths for the PyQt top level directory
        candidates = ["../", "/usr/local/", "/usr/local/src/", "/usr/src/", "/usr/share/sip", "/usr/local/share/sip"]

        for path in candidates:
            if path [-3:] == 'sip':
                pyqtsipdir = path
            else:
                pyqtsipdir = os.path.join (path, pyqt, 'sip')

            if os.access (path, os.F_OK):
                if not os.access (os.path.join (pyqtsipdir, 'qtmod.sip'), os.F_OK):
                    continue
                pyQtSipDir = pyqtsipdir
                break

    elif not (os.access (pyQtSipDir, os.F_OK) and os.access (pyQtSipDir + '/qtmod.sip', os.F_OK)):
        pyQtSipDir = None

    if not pyQtSipDir:
        error ("Couldn't find PyQt/sip directory")
    else:
        inform ("PyQt/sip directory found at %s" % pyQtSipDir)

    checkPyQtSipVersion ()
    checkSipVersion ()

def hexVersionToString (vs):
    if vs [2] != "0":
        vs = string.replace (vs, "0x", "0x0")
    major = str (int (vs [2:4]))
    minor = str (int (vs [4:6]))
    point = str (int (vs [6:8]))
    if point != 0:
        v = "%s.%s" % (major, minor)
    else:
        v = "%s.%s.%s" % (major, minor, point)

    return v

def checkPyQtSipVersion ():
    """
    Check the version of the PyQt sip files - snapshots should work here
    too (PYQT_VERSION_STR will have the snapshot name but PYQT_VERSION
    retains the version number of the previous version)
    """
    f = open (os.path.join (pyQtSipDir, "qtmod.sip"), "r")
    for line in f.readlines ():
        if string.find (line, "static unsigned PYQT_VERSION") >= 0:
            vs    = string.strip (string.split (line, "=") [1])
            pyqtv = hexVersionToString (vs)
            if pyqtv == Release:
                inform ("PyQt sip file version is compatible with PyKDE (%s)" % (Release))
                return
            else:
                error ("PyQt %s sip files not compatible with PyKDE %s" % (pyqtv, Release))

    else:
        error ("Unable to detrermine PyQt sip file version")


def checkSipVersion ():
    """
    This checks the version of the sip *file generator* but not
    the version of the sip libs - hopefully they're the same
    """

    if os.system ("sip -V > sipv.tmp") or not os.access ("sipv.tmp", os.F_OK):
        error ("Error trying to determine sip version")

    f = open ("sipv.tmp", "r")
    buff = f.read ()
    f.close ()
    os.remove ("sipv.tmp")

    if buff [3] == ".":
        sipv = buff [0:5]
    else:
        sipv = buff [0:3]

    if sipv != Release:
        from qt import PYQT_VERSION_STR
        if string.find (buff, PYQT_VERSION_STR) != 0:
            error ("sip version %s not compatible with PyKDE %s" % (sipv, Release))

    inform ("sip version is compatible with PyKDE-%s" % (Release))


def installChecks():
    """Carry out basic checks about the installation.
    """
    inform("Building PyKDE %s for Python %s on %s." % (Release, pyFullVers,sys.platform))

    # Check we are in the top level directory.
    if not os.access(progName,os.F_OK):
        error("This program must be run from the top level directory of the package, ie. the directory containing this program")

    if not os.access(libDir, os.F_OK) \
        and runProgram (platMkdir, [libDir]) \
        and not os.access(libDir,os.F_OK):
        error("The %s temporary library destination directory cannot be created." % (libDir))

    # Check the installation directory is valid.
    if not os.access(modDir,os.F_OK) \
        and runProgram (platMkdir, [modDir]) \
        and not os.access(modDir,os.F_OK):
        error("The %s PyKDE destination directory cannot be created. Use the -d argument to set the correct directory or check permissions." % (modDir))

    proPatches["DESTDIR"] = [re.compile("@BL_DESTDIR@",re.M), os.path.join ("..", libDir)]

    inform("%s is the PyKDE installation directory." % (modDir))

    # Check that the Python header files have been installed.
    if not os.access(platPyIncDir + "/Python.h",os.F_OK):
        error("Python.h could not be found in %s. Make sure Python is fully installed, including any development packages" % (platPyIncDir))

    inform("%s contains Python.h." % (platPyIncDir))

    # Look for the SIP binary.
    global sipBin

    if sipBin is None:
        sipBin = searchPath("sip",platBinDir)
    elif not isProgram(sipBin):
        sipBin = None

    if sipBin is None:
        error("The SIP code generator could not be found. Use the -p argument to set the correct program")

    inform("%s will be used as the SIP code generator." % (sipBin))

    # Check that the SIP header files have been installed.
    if not os.access(sipIncDir + "/sipQt.h",os.F_OK):
        error("sipQt.h could not be found in %s. Use the -e argument to set the correct directory" % (sipIncDir))

    inform("%s contains sipQt.h." % (sipIncDir))

    global sipModuleDir

    if sipModuleDir is None:
        dirlist = [platPyDLLDir, platPyScriptDir, platPySiteDir]
    else:
        dirlist = [sipModuleDir]
        sipModuleDir = None

    for d in dirlist:
        if glob.glob(d + "/libsip*"):
            sipModuleDir = d
            break

    if sipModuleDir is None:
        error("The SIP module/library could not be found. Use the -s argument to set the correct directory")

    proPatches["LIBS"] = [re.compile("@BL_SIPMODDIR@",re.M), "%s -L %s" % (os.path.join ("..", libDir), sipModuleDir)]

    inform("%s contains the SIP module." % (sipModuleDir))

    # Check the Qt installation.
    checkQtDirAndVersion()
    checkKDEDirAndVersion()

    # Look for the Makefile generator.
    checkMakefileGenerator()

    defines = ""

    proPatches["DEFINES"] = [re.compile("@BL_DEFINES@",re.M), defines]

    checkQtLibrary()
    checkThreading()
    checkKDELibrary()
    checkPyQt()

    # Look for Make.
    global makeBin

    if makeBin is None:
        makeBin = searchPath(platMake)
    elif not isProgram(makeBin):
        makeBin = None

    if makeBin is None:
        error("A make program could not be found. Use the -m argument to set the correct program")

    # For lib template project files.
    makefilePatches["MAKE_LIB"] = [re.compile("(^CC)",re.M), "MAKE     = \"" + makeBin + "\"\n\\1"]

    # For subdir template project files.
    makefilePatches["MAKE_SUBDIR"] = [re.compile("(^MAKEFILE)",re.M), "MAKE     = \"" + makeBin + "\"\n\\1"]

    inform("%s will be used as the make program." % (makeBin))


def checkMakefileGenerator():
    """Check that the Makefile generator can be found.
    """
    global makefileGen

    if makefileGen is None:
        mfg = "qmake"
        xtradir = qtDir + "/bin"
        makefileGen = searchPath(mfg,xtradir)
    elif not isProgram(makefileGen):
        mfg = makefileGen
        makefileGen = None

    if makefileGen is None:
        error("The Makefile generator %s could not be found. Use the -g argument to set the correct program" % (mfg))

    # Make sure that references to qmake use the correct absolute pathname.
    makefilePatches["QMAKE1"] = [re.compile("(^QMAKE[ \t]*=).*$",re.M), "\\1 \"" + makefileGen + "\""]

    # Remove the invocation of the qmake_all target.
    makefilePatches["QMAKE2"] = [re.compile(" qmake_all",re.M), ""]

    inform("%s will be used to generate Makefiles." % (makefileGen))


def compileChecks():
    """Tune the compiler flags, and see which PyQt modules to build.
    """
    # See if we can re-run make with some preferred options.  Don't do it for
    # Windows because we know the options aren't supported - and MSVC only
    # issues a warning so we can't detect the result properly anyway.
    if sys.platform != "win32":
        copyToFile("qttest.cpp",
"""int main(int argc,char **argv)
{
    return 0;
}
""")

        inform("Checking to see if the C++ compiler supports -fno-exceptions.")

        tmpPatches = {"CXXFLAGS": [re.compile("(^CXXFLAGS.*)$",re.M), "\\1 -fno-exceptions\n"]}
        patchFile("Makefile",tmpPatches,"Makefile.bak")

        if runMake(None,0):
            # Restore the old working Makefile.
            os.remove("Makefile")
            os.rename("Makefile.bak","Makefile")

            inform("The C++ compiler doesn't support -fno-exceptions.")
        else:
            os.remove("Makefile.bak")

            makefilePatches["CXXFLAGS"] = [re.compile("(^CXXFLAGS.*)$",re.M), "\\1 -fno-exceptions\n"]

            inform("The C++ compiler supports -fno-exceptions.")


def generateFeatures(featfile):
    """Generate the header file describing the Qt features that are enabled if
    it doesn't already exist.  (If it already exists then we are probably cross
    compiling and generated the file through other means.)

    featfile is the name of the features file.
    """
    if os.access(featfile,os.F_OK):
        inform("Using existing features file.")
        return

    runMake("clean")

    inform("Generating the features file.")

    # The features that a given Qt configuration may or may not support.
    flist = ["ACTION", "CLIPBOARD", "CODECS", "COLORDIALOG", "DIAL", "DNS",
             "DOM", "DRAGANDDROP", "ICONVIEW", "IMAGE_TEXT", "INPUTDIALOG",
             "FILEDIALOG", "FONTDATABASE", "FONTDIALOG", "MESSAGEBOX",
             "MIMECLIPBOARD", "NETWORKPROTOCOL", "PICTURE", "PRINTDIALOG",
             "PRINTER", "PROGRESSDIALOG", "PROPERTIES", "SEMIMODAL",
             "SIZEGRIP", "SOUND", "SPLITTER", "STYLE_CDE", "STYLE_INTERLACE",
             "STYLE_MOTIF", "STYLE_MOTIFPLUS", "STYLE_PLATINUM", "STYLE_SGI",
             "STYLE_WINDOWS", "TABDIALOG", "TABLE", "TABLEVIEW",
             "TRANSFORMATIONS", "TRANSLATION", "WIZARD", "WORKSPACE"]

    # Generate the program which will generate the features file.
    f = open("qttest.cpp","w")

    f.write(
"""#include <stdio.h>
#include <Python.h>
#include <qglobal.h>
#include <qapplication.h>

int main(int argc,char **argv)
{
    FILE *fp;
    QApplication app(argc,argv,0);

    if ((fp = fopen(\"%s\",\"w\")) == NULL)
        return 1;

#if !defined(QT_THREAD_SUPPORT) || !defined(WITH_THREAD)
    fprintf(fp,"-x Qt_THREAD_SUPPORT\\n");
#endif
""" % (featfile))

    for feat in flist:
        f.write(
"""
#if defined(QT_NO_%s)
    fprintf(fp,"-x Qt_%s\\n");
#endif
""" % (feat,feat))

    f.write(
"""
    fclose(fp);

    return 0;
}
""")

    f.close()

    runMake()
    os.system ("pwd")
    runProgram("./qttest")

    inform("Generated the features file.")


def addBuildDir(file,wd):
    """Add an optional build path to a given file.

    file is the file name.
    wd is an optional build directory.

    returns the file name.
    """
    if wd:
        file = os.path.join (wd, file)
    return file


def generateSource(mname,plattag,qttag,xtrtag=None,cpu_pool=None):
    """Generate the C++ source code for a particular PyQt module.

    mname is the name of the module.
    plattag is the SIP tag for the platform.
    qttag is the SIP tag for the Qt version.
    xtrtag is an optional extra SIP tag.
    cpu_pool is an optional bounded semaphore to control threading.
    """
    if cpu_pool:
        cpu_pool.acquire()

    inform("Generating the C++ source for the %s module." % (mname))

    libWC = "lib%s*" % (mname)

    try:
        shutil.rmtree(mname)
        if runProgram (platRmdir, [os.path.join ("..", libDir, libWC)]) != 0:
            error ("Unable to delete old library files in %s" % (libDir))
    except:
        pass

    try:
        os.mkdir(mname)
    except:
        error("Unable to create the %s directory" % (mname))

    pro = mname + ".pro"
    fullmname = mname + "-kde" + str(kdeVersion) + ".sip"

    kdetag = makeVersion ('KDE', str (kdeVersion))

    # plattag          is Windows, Linux (X11), MacOS (>)
    # qttag and kdetag are Qt and KDE versions
    # features         is the features file
    # sip              is the source (sip) file subdirectory
    # pyQtSipDir       is where the PyQt sip files are
    # mname            is the module name (-m for generating
    #                  the make files, -c for the actual module name)
    # fullname         is the name of the top level sip file
    #                  that controls sip file inclusion, imports, etc.
    argv = ["-t", plattag,
            "-t", qttag,
            "-t", kdetag,
            "-z", "features",
            "-I", "sip",
            "-I", pyQtSipDir,
            "-m", mname + "/" + pro,
            "-c", mname,
            fullmname]

    if xtrtag:
        argv.insert(0,xtrtag)
        argv.insert(0,"-t")

#    print qttag, kdetag

    runProgram (sipBin, argv)
    runProgram (sys.executable + " postproc", ['-p', mname, "-o", "tr", "*.cpp"])
    runProgram (sys.executable + " postproc", ['-p', mname, "-o", "varhier", "*.cpp"])

    if mname == "kdeui":
        runProgram (sys.executable + " postproc", ["-p ", mname, "-o", "shpix",  "sipkdeuiKSharedPixmap.cpp"])

    # Python names binary modules differently on different platforms.
    target = mname + "cmodule"

    # Generate the Makefile.
    inform("Generating the Makefile for the %s module." % (mname))

    # deep copy the patch dicts for threading safety
    _makepat = makefilePatches.copy()
    _propat = proPatches.copy()

    _propat["TARGET"] = [re.compile("@BL_TARGET@",re.M), target]

    if catCppFiles:
        catFiles(mname,mname) # module, wd

    # creates the make files
    buildMakefile(pro,_propat,_makepat,mname)

    # edits the make files
    fixInstallTarget(mname)

    # Compile the Python part of the module.
    pyname = mname + ".py"

    inform("Compiling %s." % (pyname))
    py_compile.compile(addBuildDir(pyname,mname))

    if cpu_pool:
        cpu_pool.release()


def pushDir(newdir):
    """Change to a new directory and return the old one.

    Note: using this function is not thread safe!

    newdir is the new directory.
    """
    olddir = os.getcwd()
    os.chdir(newdir)

    return olddir


def popDir(olddir):
    """Change back to a previous directory.

    Note: using this function is not thread safe!

    olddir is the old directory.
    """
    os.chdir(olddir)


def runProgram(prog,argv = [],fatal = 1,wd = None):
    """Execute a program.

    prog is the name of the program.
    argv is the optional list of program arguments.
    fatal is non-zero if an error is considered fatal.
    wd is an optional build directory.

    Returns the error code.
    """
    try:
        run = prog + ' ' + string.join(argv)
        if wd:
            # FIXME: reason, why theading will not work on win32 yet <hpj@urpla.net>
            run = 'cd ' + wd + ' >/dev/null && ' + run
        #inform("run: <%s>" % run)
        rc = os.system (run)
    except:
        raise

    if rc != 0 and fatal:
        raise
        error("%s failed with an exit code of %d." % (prog,rc))

    return rc


def runMake(target = None,fatal = 1):
    """Run make.

    target is the optional make target.
    fatal is non-zero if an error is considered fatal.

    Returns the error code.
    """
    if target is None:
        argv = []
    else:
        argv = [target]

    return runProgram(makeBin,argv,fatal)


def buildMakefile(pro,_propat = None,_makepat = None,wd = None):
    """Run qmake or tmake to create a Makefile from a project file.

    pro is the name of the project file.
    _propat is an optional project patch dictionary.
    _makepat is an optional project patch dictionary.
    wd is an optional build directory.
    """
    if not _propat:
        _propat = proPatches
    if not _makepat:
        _makepat = makefilePatches

    # Preserve the original project file.
    pf = addBuildDir(pro,wd)
    probak = pf + ".bak"
    patchFile(pf,_propat,probak)

    # Make sure that the Makefile has a later timestamp than the project file.
    time.sleep(1)

    runProgram(makefileGen,['-o', 'Makefile', pro],1,wd)
    os.remove(pf)
    os.rename(probak,pf)

    patchFile(addBuildDir("Makefile",wd),_makepat)


def catFiles(mname,wd = None):
    """Concatenate a modules C++ source files.

    pro is the name of the project file.
    wd is an optional build directory.
    """
    am = ""
    if catSplit > 1:
        am = " (into %d pieces)" % catSplit
    inform("Concatenating the C++ files for the %s module%s." % (mname, am))

    pro = addBuildDir(mname + ".pro",wd)

    f = open(pro,"r")
    buf = f.read()
    f.close()

    pat = re.compile("^SOURCES =.*.cpp\n",re.M | re.S)
    match = re.search(pat,buf)
    srclist = buf[match.start(0):match.end(0)]
    srclist = srclist[13:-1]
    srclist = string.replace(srclist,"\\\n\t","")
    srclist = string.split(srclist," ")

    # Concatenate the files.
    p = 0
    plist = []
    nrinpiece = (len(srclist) + catSplit - 1) / catSplit

    #inform("Combining %d modules into %d pieces (%d)" % (len(srclist), catSplit, nrinpiece))

    # Do each piece.
    while srclist:
        pname = "%shuge%d.cpp" % (mname, p)
        plist.append(pname)

        d = open(addBuildDir(pname,wd),"w")

        for cppfile in srclist[:nrinpiece]:
            f = open(addBuildDir(cppfile,wd),"r")
            d.write(f.read())
            f.close()

        d.close()

        srclist = srclist[nrinpiece:]
        p = p + 1

    # Replace the C++ file names in the project file.
    buf = re.sub(pat,"SOURCES = " + string.join(plist," ") + "\n",buf)

    f = open(pro + ".new","w")
    f.write(buf)
    f.close()

    os.remove(pro)
    os.rename(pro + ".new",pro)


def fixInstallTarget(mname = None):
    """Fix the install target for a Makefile.

    mname is the name of the PyQt module, or None if the target isn't for a
    module.

    """
    
    if not mname:
        return

    # The correct path for use in the install commands
    # is <module name>/Makefile - the previous version
    # was omitting the module name, so all mods ended
    # up being done to the top level Makefile
    fn    = os.path.join (mname, "Makefile")
    newFn = os.path.join (mname, "Makefile.new")
    libWC = "lib%s*" % (mname)

    fi     = open(fn,"r")
    fo     = open(newFn,"w")
    buff  = fi.read ()

    installTxt = ("\ninstall:\n")

    # <module name>y and .pyc are created in the <module name> dir
    # (eg dcop.py is created in ../dcop)
    # All libs are written to ,,/libs - this code just causes
    # 'make install' to copy all of the relevant files to the
    # correct end destination (normally python/site-packages)
    # This allows 'make' to be run as a user but 'make install'
    # still requires root accesss, at least for site-packages
    # (as it should)
    installTxt += "\t-%s %s.py %s\n" % (platCopy,mname,modDir)\
                + "\t-%s %s.pyc %s\n" % (platCopy,mname,modDir)\
                +"\t-%s %s %s\n" % (platMove, os.path.join ("..", libDir, libWC), modDir)

    cleanTxt    = "clean:\n\t-%s %s\n\t-" % (platRemove, os.path.join ("..", libDir, libWC))

    buff = re.compile ("^clean: ", re.M).sub (cleanTxt, buff)
    buff = re.compile ("install: all").sub (installTxt, buff)

    fo.write (buff)
    fi.close()
    fo.close()
    os.unlink (fn)
    os.rename (newFn, fn)


def versionToTag(vers,tags,desc):
    """Convert a version number to a tag.

    vers is the version number.
    tags is the dictionary of tags keyed by version number.
    desc is the descriptive name of the package.

    Returns the corresponding tag.
    """
    tag = None

    vl = tags.keys()
    vl.sort()

    # For a snapshot use the latest tag.
    if vers == 0:
        tag = tags[vl[-1]]
    else:
        for v in vl:
            if vers < v:
                tag = tags[v]
                break

        if tag is None:
            error("Invalid %s version: 0x%06x." % (desc,vers))

    return tag


def fileOpts (fn):
    try:
        optfile = open (fn, 'r')
    except:
        error ("Could not open option file %s" % (fn))

    opts = []

    for line in optfile.readlines ():
        if (line [0] == '#') or (line == '\n'):
           continue
        elif line [0] == '-':
            opts.append ((line [0:2], string.strip (line [2:])))
        else:
            opts.append ((line [0:1], string.strip (line [1:])))

    print 'Additional options: ',
    for opt, arg in opts:
        print opt, arg,
    print

    return opts


def main(argv):
    """The main function of the script.

    argv is the list of command line arguments.
    """
    import getopt

    # Parse the command line.
    global progName
    progName = os.path.basename(argv[0])

    # check, if threading is desired.
    try:
        nrcpu = int(os.environ["SMP"])
    except:
        nrcpu = 1

    # check, if threading is available.
    if nrcpu > 1:
        try:
            import signal   # should improve signal distribution with threads
            from threading import BoundedSemaphore, Thread
        except:
            inform("Threaded build is not available")
        else:
            global nrThreads
            nrThreads = nrcpu

    initGlobals()

    try:
        optlist, args = getopt.getopt(argv[1:],"hc:d:e:g:i:j:k:l:m:n:o:p:q:r:s:t:v:z:")
    except getopt.GetoptError:
        usage()

    # Look for '-z' first and process that switch
    # (command line switches override file switches)
    for opt, arg in optlist:
        if opt == "-z":
             optlist = fileOpts (arg) + optlist

    explicitMake = 0

    for opt, arg in optlist:
        if opt == "-h":
            usage(0)
        elif opt == "-c":
            global catCppFiles
            if arg == "-":
                catCppFiles = 0
            elif arg == "+":
                catCppFiles = 1
            else:
                usage ()
        elif opt == "-d":
            global modDir
            modDir = arg
        elif opt == "-e":
            global sipIncDir
            sipIncDir = arg
        elif opt == "-g":
            global makefileGen
            makefileGen = arg
        elif opt == "-i":
            global qtIncDir
            qtIncDir = arg
        elif opt == "-j":
            global catSplit
            try:
                catSplit = int(arg)
            except:
                catSplit = 0
            if catSplit < 1:
                usage()
        elif opt == "-k":
            global kdeDir
            kdeDir = arg
        elif opt == "-l":
            global qtLib
            if arg in ("qt", "qt-mt"):
                qtLib = arg
            else:
                usage()
        elif opt == "-m":
            global makeBin
            makeBin = arg
            explicitMake = 1
        elif opt == "-n":
            global kdeIncDir
            kdeIncDir = arg
        elif opt == "-o":
            global kdeLib
            kdeLib = arg
        elif opt == "-p":
            global sipBin
            sipBin = arg
        elif opt == "-q":
            global qtDir
            qtDir = arg
        elif opt == "-r":
            global qtLibDir
            qtLibDir = arg
        elif opt == "-s":
            global sipModuleDir
            sipModuleDir = arg
        elif opt == "-t":
            global platPyLibDir
            platPyLibDir = arg
        elif opt == "-v":
            global pyQtSipDir
            pyQtSipDir = arg
        elif opt == "-z":
            pass

    initGlobals()

    installChecks()
    maindir = mkTempBuildDir()
    compileChecks()

    # Work out the platform and Qt version tags to pass to SIP to generate the
    # code we need.
    if qtLib == "qte":
        plattag = "WS_QWS"
    elif sys.platform == "win32":
        plattag = "WS_WIN"
    else:
        plattag = "WS_X11"


    qttag = versionToTag(qtVersion,qttags,"Qt")

    # Generate the features file.
    generateFeatures(os.path.join (maindir, "features"))

    # We don't need the temporary build directory anymore.
    mkTempBuildDir(maindir)

    # Add build dir as additional -L target
    proPatches["BUILDDIR"] = [re.compile("@BL_BUILDDIR@",re.M), modDir]

    # generate the source modules
    buildModules = modList [str (kdeVersion) [0:2]]

    # prepare threads
    if nrThreads:
        threads = []
        cpu_pool = BoundedSemaphore(nrThreads)
        inform("Starting %d builder threads for %s" % \
                (nrThreads, string.join(buildModules, " ")))

    subdirs = ""
    for mname in buildModules:
        subdirs = subdirs + " " + mname
        if nrThreads:
            t = Thread(target=generateSource, name=mname,
                        args=(mname, plattag, qttag, None, cpu_pool))
            threads.append(t)
            t.start()
        else:
            generateSource(mname, plattag, qttag)

    # wait for threads to finish..
    if nrThreads:
        while 1:
            for t in threads:
                if t.isAlive():
                    t.join(1.0)
                else:
                    del threads[threads.index(t)]
            if not threads:
                break
        inform("Builder threads finished.")

    # Generate the top-level Makefile.
    inform("Creating top level Makefile.")
    copyToFile("PyKDE.pro","TEMPLATE = subdirs\nSUBDIRS =" + subdirs + "\n")
    buildMakefile("PyKDE.pro")

    # Tell the user what to do next.
    if explicitMake:
        mk = makeBin
    else:
        (mk,ignore) = os.path.splitext(os.path.basename(makeBin))

    msg =\
"""

The build of the PyKDE source code for your system is now complete. To compile and install
PyKDE run \"%s\" as a user and \"%s install\" with privileges appropriate to the destination
directory
""" % (mk,mk)

    inform(msg)


if __name__ == "__main__":
    try:
        main(sys.argv)
    except SystemExit:
        raise
    except:
        print internalError
        raise
